from pysymbolscanner.utils import filter_duplicate_dicts_in_list


class Stock:
    def __init__(
        self,
        name,
        wiki_name,
        link,
        link_lang,
        main_symbol,
        loc,
        indices,
        industry,
        symbols,
        isins,
        founded,
        employees,
    ):
        self.data = {
            'name': name,
            'wiki_name': wiki_name,
            'link': link,
            'link_lang': link_lang,
            'symbol': main_symbol,
            'country': loc,
            'indices': indices,
            'industries': industry,
            'symbols': symbols,
            'isins': isins,
            'metadata': {'founded': founded, 'employees': employees},
        }

    @classmethod
    def from_wiki(cls, indices, wiki_name, link, link_lang, symbol):
        return cls(
            wiki_name,
            wiki_name,
            link,
            link_lang,
            symbol,
            '',
            indices,
            [],
            [],
            [],
            0,
            0,
        )

    def to_pyticker_symbol(self):
        sym = self.symbol
        symbols = self.get_symbols(self.symbols)
        if self.yahoo_symbols:
            symbols += self.yahoo_symbols
            symbols = filter_duplicate_dicts_in_list(symbols)

        if not sym:
            sym = next(map(lambda x: x['yahoo'], symbols), None)

        if sym and '.' in sym:
            sym = sym.split('.')[0]

        return {
            'name': self.name,
            'wiki_name': self.wiki_name,
            'symbol': sym if sym else self.get_symbol(self.symbols),
            'country': self.country,
            'indices': self.indices,
            'industries': self.industries,
            'symbols': symbols,
            'isins': self.isins,
            'metadata': self.metadata,
        }

    @property
    def name(self):
        return self.data['name']

    @name.setter
    def name(self, name):
        self.data['name'] = name

    @property
    def wiki_name(self):
        return self.data['wiki_name']

    @wiki_name.setter
    def wiki_name(self, wiki_name):
        self.data['wiki_name'] = wiki_name

    @property
    def symbol(self):
        return self.data['symbol']

    @symbol.setter
    def symbol(self, symbol):
        self.data['symbol'] = symbol

    @property
    def yahoo_symbols(self):
        return self.data['yahoo_symbols']

    @yahoo_symbols.setter
    def yahoo_symbols(self, symbols):
        self.data['yahoo_symbols'] = symbols

    @property
    def link(self):
        return self.data['link']

    @link.setter
    def link(self, link):
        self.data['link'] = link

    @property
    def link_lang(self):
        return self.data['link_lang']

    @link_lang.setter
    def link_lang(self, lang):
        self.data['link_lang'] = lang

    @property
    def country(self):
        return self.data['country']

    @country.setter
    def country(self, country):
        self.data['country'] = country

    @property
    def industries(self):
        return self.data['industries']

    @industries.setter
    def industries(self, industries):
        self.data['industries'] = industries

    @property
    def indices(self):
        return self.data['indices']

    @indices.setter
    def indices(self, indices):
        self.data['indices'] = indices

    @property
    def symbols(self):
        return self.data['symbols']

    @symbols.setter
    def symbols(self, symbols):
        self.data['symbols'] = symbols

    @property
    def isins(self):
        return self.data['isins']

    @isins.setter
    def isins(self, isins):
        self.data['isins'] = isins

    @property
    def metadata(self):
        return self.data['metadata']

    @metadata.setter
    def metadata(self, metadata):
        self.data['metadata'] = metadata

    @property
    def metadata_founded(self):
        return self.data['metadata']['founded']

    @metadata_founded.setter
    def metadata_founded(self, founded):
        if not isinstance(founded, int):
            raise ValueError('founded must be of type int')
        self.data['metadata']['founded'] = founded

    @property
    def metadata_employees(self):
        return self.data['metadata']['employees']

    @metadata_employees.setter
    def metadata_employees(self, employees):
        if not isinstance(employees, int):
            raise ValueError('employees must be of type int')
        self.data['metadata']['employees'] = employees

    @staticmethod
    def get_symbol(symbols):
        if not symbols:
            return None
        return min(map(lambda x: x.split('|')[-1], symbols))

    @staticmethod
    def get_symbols(symbols):
        if not symbols:
            return []
        result = []
        for exchange, sym in filter(
            lambda x: len(x) == 2, map(lambda x: x.split('|'), symbols)
        ):
            exc = exchange.upper().strip()
            if exc == 'FWB':
                result.append(
                    {
                        'yahoo': f'{sym}.F',
                        'google': f'FRA:{sym}',
                        'currency': 'EUR',
                    }
                )
            elif exc in ['NASDAQ', 'NASDAQSYMBOL']:
                result.append(
                    {
                        'yahoo': sym,
                        'google': f'NASDAQ:{sym}',
                        'currency': 'USD',
                    }
                )
            elif exc in ['NYSE', 'NEW YORK STOCK EXCHANGE']:
                result.append(
                    {'yahoo': sym, 'google': f'NYSE:{sym}', 'currency': 'USD'}
                )
            elif exc == 'Euronext':
                continue
            elif exc == 'LSE':
                continue
            elif exc == 'ASX':
                continue
            elif exc in ['BMAD', 'BMV']:
                continue
            elif exc == 'ISE':
                continue
            elif exc == 'MICEX-RTS':
                result.append(
                    {
                        'yahoo': f'{sym}.ME',
                        'google': f'MCX:{sym}',
                        'currency': 'RUB',
                    }
                )
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" pysymbolscanner
  Copyright 2020 Slash Gordon
  Use of this source code is governed by an MIT-style license that
  can be found in the LICENSE file.
"""

import logging
import multiprocessing
import os
import pickle
import sys

import pandas as pd
import requests
from bs4 import BeautifulSoup
from pytickersymbols import PyTickerSymbols

from pysymbolscanner.const import most_common_endings
from pysymbolscanner.index_definitions import Indices
from pysymbolscanner.stock import Stock
from pysymbolscanner.wiki import get_merged_infobox, get_wiki_url
from pysymbolscanner.word_score import deep_search, get_best_match
from pysymbolscanner.yahoo import YahooSearch


class SymbolScanner:
    MIN_WORD_OCCURRENCE = 5
    MAX_PROCESSES = 8
    PICKLE_FILE = 'index_data.pickle'

    def __init__(self, cache):
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
        self.log = logging.getLogger('symbol scanner')
        self.data = None
        self.cache = cache

    def start(self):
        if self.cache and os.path.isfile(SymbolScanner.PICKLE_FILE):
            with open(SymbolScanner.PICKLE_FILE, 'rb') as handle:
                self.data = pickle.load(handle)
        else:
            self.data = self.start_index()
            self.data = self.start_metadata()
            self.data = self.start_yahoo()
            self.data = self.start_sync()
            with open(SymbolScanner.PICKLE_FILE, 'wb') as handle:
                pickle.dump(
                    self.data, handle, protocol=pickle.HIGHEST_PROTOCOL
                )

    def start_yahoo(self):
        result = {}
        with multiprocessing.Pool(processes=self.MAX_PROCESSES * 4) as pool:
            for index in self.data:
                args = map(
                    lambda stock: stock,
                    self.data[index],
                )
                stocks_with_yahoo_symbols = pool.map(self.worker_yahoo, args)
                result[index] = stocks_with_yahoo_symbols
        return result

    def worker_yahoo(self, stock):
        search = YahooSearch()
        if not stock.name:
            self.log.warn(
                f'Item without name {stock}'
            )
            return stock
        symbols = search.get_symbols(stock.name)
        stock.yahoo_symbols = symbols
        return stock

    def start_sync(self):
        stocks = PyTickerSymbols()
        if not self.data:
            return

        stocks_py = stocks.get_all_stocks()

        names_py = list(set([stock['name'] for stock in stocks_py]))

        names_wiki = list(
            set(
                [
                    stock.wiki_name
                    for stock_list in map(
                        lambda index: self.data[index],
                        Indices.symbol_source_dict,
                    )
                    for stock in stock_list
                ]
            )
        )

        names = names_py + names_wiki
        endings = list(
            map(lambda x: ' ' + x, self.get_most_common_endings(names))
        )
        occurrences = list(
            set(
                list(self.get_most_common_occurrences(names))
                + most_common_endings
                + endings
            )
        )

        occurrences.sort(key=len, reverse=True)
        endings.sort(key=len, reverse=True)
        with multiprocessing.Pool(processes=self.MAX_PROCESSES) as pool:
            items = list(
                map(
                    lambda index: (index, endings, occurrences, stocks_py),
                    Indices.symbol_source_dict,
                )
            )
            sync_results = pool.map(self.worker_sync, items)
            result = dict(map(lambda val: list(val.items())[0], sync_results))
            return result

    @staticmethod
    def find_by_symbol(stock, py_stocks):
        for idx, py_stock in enumerate(py_stocks):
            for py_symbol in py_stock['symbols']:
                for symbol in stock['symbols']:
                    if symbol['yahoo'] == py_symbol['yahoo']:
                        return idx
        return -1

    @staticmethod
    def find_by_symbol_and_country(stock, py_stocks):
        for idx, py_stock in enumerate(py_stocks):
            if (
                stock['symbol'] == py_stock['symbol']
                and stock['country'] == py_stock['country']
                and stock['metadata'] == py_stock['metadata']
            ):
                return idx
        return -1

    def worker_sync(self, args):
        index, endings, occurrences, stocks = args
        wiki_stocks = self.data[index].copy()
        for idx, wiki_stock in enumerate(wiki_stocks):
            stock = wiki_stock.to_pyticker_symbol()
            names = [stock['name'] for stock in stocks]
            name_id = self.find_by_symbol(stock, stocks)
            if name_id != -1:
                wiki_stocks[idx].name = names[name_id]
                continue

            name_id = self.find_by_symbol_and_country(stock, stocks)
            if name_id != -1:
                wiki_stocks[idx].name = names[name_id]
                continue

            wiki_stock_name = stock['wiki_name']
            name_id, max_score = get_best_match(
                wiki_stock_name,
                names,
                word_filter=occurrences,
            )
            if max_score > 0.9:
                self.log.warn(
                    f'Sync by word score. {wiki_stocks[idx].wiki_name} =='
                    f' {names[name_id]} = {max_score}'
                )
                wiki_stocks[idx].name = names[name_id]
                continue

            name_id, max_score = deep_search(
                wiki_stock_name, names, occurrences
            )

            if max_score > 0.9:
                self.log.warn(
                    f'Sync by deep search. {wiki_stocks[idx].wiki_name} =='
                    f' {names[name_id]} = {max_score}'
                )
                wiki_stocks[idx].name = names[name_id]
            else:
                self.log.warn(
                    f'Did not find any match for {index} - {wiki_stock_name}'
                )
                wiki_stocks[idx].name = wiki_stocks[idx].wiki_name
        return {index: wiki_stocks}

    def get_most_common_endings(self, names):
        most_words = {}
        for name in names:
            if not name.split():
                continue
            word = name.split()[-1]
            most_words[word] = most_words.get(word, 0) + 1

        items_list = [word for word in most_words.items()]
        items_list.sort(key=lambda tup: tup[1])
        return map(
            lambda x: x[0],
            filter(lambda f: f[1] >= self.MIN_WORD_OCCURRENCE, items_list),
        )

    def get_most_common_occurrences(self, names):
        most_words = {}
        for name in names:
            if not name.split():
                continue
            words = name.split()
            for word in words:
                most_words[word] = most_words.get(word, 0) + 1

        items_list = [word for word in most_words.items()]
        items_list.sort(key=lambda tup: tup[1])
        return map(
            lambda x: x[0],
            filter(lambda f: f[1] >= self.MIN_WORD_OCCURRENCE, items_list),
        )

    def start_metadata(self):
        result = {}
        with multiprocessing.Pool(processes=self.MAX_PROCESSES * 4) as pool:
            for index in self.data:
                args = map(
                    lambda stock, idx=index: (
                        stock,
                        Indices.symbol_source_dict[idx].PAGE_LANGS,
                    ),
                    self.data[index],
                )
                stocks_with_metadata = pool.map(self.worker_metadata, args)
                result[index] = stocks_with_metadata
        return result

    def worker_metadata(self, args):
        stock, langs = args
        infobox = get_merged_infobox(
            stock.wiki_name, stock.link, stock.link_lang, langs
        )
        if not infobox:
            self.log.warn(
                f'Did not find any wikipedia data for {stock.wiki_name}'
            )
            return stock
        return infobox.to_stock(stock.indices, stock.symbol)

    def start_index(self):
        with multiprocessing.Pool(processes=self.MAX_PROCESSES) as pool:
            items = list(
                map(lambda x: (x[0], x[1]), Indices.symbol_source_dict.items())
            )
            index_results = pool.map(self.worker_index, items)
            result = dict(map(lambda val: list(val.items())[0], index_results))
            return result

    def worker_index(self, args):
        key, index_source = args
        self.log.info(f'________________ parse {index_source.TITLE}')
        wiki_url = get_wiki_url(index_source.LANG, index_source.TITLE)
        response = requests.get(wiki_url)
        dfs = pd.read_html(response.text)
        table_id = self.get_table_id(dfs, index_source.ITEMS)
        if table_id == -1:
            raise RuntimeError(f'Could not find index table for {key}')
        df = dfs[table_id]
        try:
            links = self.get_wiki_links(
                response.text, table_id, index_source.COL_NAME_COMPANY
            )
        except TypeError:
            raise RuntimeError(f'Could not find company links {key} table')
        df['link'] = links
        if index_source.COL_NAME_SYMBOL is not None:
            df = df.rename(
                columns={
                    index_source.COL_NAME_COMPANY: 'name',
                    index_source.COL_NAME_SYMBOL: 'symbol',
                }
            )[['name', 'symbol', 'link']]
        else:
            df = df.rename(columns={index_source.COL_NAME_COMPANY: 'name'})
            df['symbol'] = None
            df = df[['name', 'symbol', 'link']]
        result = {key: []}
        for wiki_name, symbol, link in zip(df.name, df.symbol, df.link):
            stock = Stock.from_wiki(
                [key], wiki_name, link, index_source.LANG, symbol
            )
            result[key].append(stock)
        return result

    @staticmethod
    def get_table_id(dfs, columns):
        table_id = -1
        for idx, df in enumerate(dfs):
            if all(col in df.columns for col in columns):
                if table_id != -1:
                    raise RuntimeError(
                        'Could not detect table because of multiple matches'
                    )
                table_id = idx
        return table_id

    @staticmethod
    def get_wiki_links(html, table_id, col_name):
        soup = BeautifulSoup(html, 'html.parser')
        table = soup.findAll('table')[table_id]
        links = []
        index = -1
        for row in table.findAll('tr'):
            column_names = row.findAll('th')
            if index == -1:
                index = SymbolScanner._get_col_id(col_name, column_names)
                continue
            tr = row.findAll('td')[index]
            try:
                link = tr.find('a')['href']
                if 'action=edit' not in link:
                    links.append(link)
                else:
                    links.append('')
            except (KeyError, TypeError):
                links.append('')
        return links

    @staticmethod
    def _get_col_id(col_name, column_names):
        index = next(
            iter(
                [
                    i
                    for i, s in enumerate(column_names)
                    if col_name in s.getText()
                ]
            ),
            None,
        )
        return index



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" pysymbolscanner
  Copyright 2020 Slash Gordon
  Use of this source code is governed by an MIT-style license that
  can be found in the LICENSE file.
"""


def filter_duplicate_dicts_in_list(items):
    return [dict(t) for t in {tuple(d.items()) for d in items}]


def flat_list(my_list):
    return [item for sublist in my_list for item in sublist]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" pysymbolscanner
  Copyright 2020 Slash Gordon
  Use of this source code is governed by an MIT-style license that
  can be found in the LICENSE file.
"""
import sys
from pysymbolscanner.command_line import symbolscanner_app

if __name__ == '__main__':
    sys.exit(symbolscanner_app())



from json import JSONDecodeError

from uplink import Consumer, Query, get, returns

from pysymbolscanner.const import remove_most_common_endings, long_to_short
from pysymbolscanner.word_score import get_score
from pysymbolscanner.utils import filter_duplicate_dicts_in_list


class YahooSearch(Consumer):

    BASE_URL = 'https://query1.finance.yahoo.com/'

    def __init__(self):
        super().__init__(base_url=YahooSearch.BASE_URL)

    @returns.json()
    @get('/v1/finance/search')
    def search(
        self,
        search: Query('q'),
        lang: Query('lang') = 'en-US',
        quotes_count: Query('quotesCount') = 15,
        news_count: Query('newsCount') = 0,
    ):
        """returns ticker symbols by search token"""

    @staticmethod
    def item_sort(item):
        return {
            'EUR': 1,
            'USD': 2,
            'RUB': 3,
            'GBP': 4,
        }.get(item['currency'], 99)

    def _search_symbols(self, search, isins=[]):
        search1 = search
        search2 = remove_most_common_endings(search)
        search3 = long_to_short(search)
        data = []
        for search_item in [search1, search2, search3] + isins:
            try:
                search_result = self.search(search_item)
                if 'quotes' not in search_result:
                    continue
                data += search_result['quotes']
            except JSONDecodeError:
                continue
        return data

    @staticmethod
    def _get_scored_quotes(search, data):
        search = remove_most_common_endings(search)
        # filter by word  score
        quotes = list(
            filter(
                lambda x: 'exchange' in x
                and (
                    (
                        'longname' in x
                        and get_score(
                            search, remove_most_common_endings(x['longname'])
                        )
                        > 0.8
                    )
                    or (
                        'shortname' in x
                        and get_score(
                            search, remove_most_common_endings(x['shortname'])
                        )
                        > 0.8
                    )
                ),
                data,
            )
        )
        exchanges = set(map(lambda x: x['exchange'], quotes))
        scored_quotes = list(
            map(
                lambda exchange: max(
                    filter(
                        lambda y, exc=exchange: y['exchange'] == exc,
                        quotes,
                    ),
                    key=lambda x: x['score'],
                ),
                exchanges,
            )
        )
        return scored_quotes

    @staticmethod
    def _get_symbol(sym, google_prefix, yahoo_suffix, currency):
        return {
            'yahoo': f'{sym}.{yahoo_suffix}' if yahoo_suffix else sym,
            'google': f'{google_prefix}:{sym}',
            'currency': currency,
        }

    @staticmethod
    def _get_pyticker_symbol(exchange, symbol):
        sym = symbol.split('.')[0] if symbol.split('.') else symbol
        return {
            'FRA': lambda x: YahooSearch._get_symbol(x, 'FRA', 'F', 'EUR'),
            'GER': lambda x: YahooSearch._get_symbol(x, 'FRA', 'F', 'EUR'),
            'BER': lambda x: YahooSearch._get_symbol(x, 'FRA', 'F', 'EUR'),
            'STU': lambda x: YahooSearch._get_symbol(x, 'FRA', 'F', 'EUR'),
            'HAN': lambda x: YahooSearch._get_symbol(x, 'FRA', 'F', 'EUR'),
            'MUN': lambda x: YahooSearch._get_symbol(x, 'FRA', 'F', 'EUR'),
            'AMS': lambda x: YahooSearch._get_symbol(x, 'AMS', 'AS', 'EUR'),
            'MCE': lambda x: YahooSearch._get_symbol(x, 'BME', 'MC', 'EUR'),
            'LSE': lambda x: YahooSearch._get_symbol(x, 'LON', 'L', 'GBP'),
            'NMS': lambda x: YahooSearch._get_symbol(x, 'NASDAQ', None, 'USD'),
            'PNK': lambda x: YahooSearch._get_symbol(
                x, 'OTCMKTS', None, 'USD'
            ),
            'NYQ': lambda x: YahooSearch._get_symbol(x, 'NYSE', None, 'USD'),
            'MCX': lambda x: YahooSearch._get_symbol(x, 'MCX', 'ME', 'RUB'),
        }.get(exchange, lambda x: None)(sym)

    def get_symbols(self, search, isins=[]):
        data = self._search_symbols(search, isins)

        if not data:
            return None

        scored_quotes = self._get_scored_quotes(search, data)
        result = []
        for quote in scored_quotes:
            exchange = quote.get('exchange', '')
            symbol = quote.get('symbol', '')
            if not symbol or not exchange:
                continue
            pyticker_symbol = self._get_pyticker_symbol(exchange, symbol)
            if pyticker_symbol:
                result.append(pyticker_symbol)
        result = filter_duplicate_dicts_in_list(result)
        result.sort(key=self.item_sort)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" pysymbolscanner
  Copyright 2020 Slash Gordon
  Use of this source code is governed by an MIT-style license that
  can be found in the LICENSE file.
"""
import re

most_common_endings = [
    'GmbH & Co. Kommanditgesellschaft auf Aktien',
    'Compagnie Générale des Établissements',
    'Compagnie Générale des établissements',
    ', Société Européenne',
    'Société Européenne',
    ' - ',
    'GmbH & Co. KGaA',
    'AG & Co. KGaA',
    'SE & Co. KGaA',
    'Aktiengesellschaft',
    'Аэрофлот',
    'PAO',
    'SA',
    'S.A.',
    'AG',
    'Ag',
    'ag',
    'SE',
    'S.E.',
    'plc',
    'Plc',
    'PLC',
    'GmbH',
    'KGaA',
    'Corp.',
    'Corp',
    'Co.',
    'Inc.',
    '(Class C)',
    '(Class B)',
    '(Class A)',
    '(A)',
    '(B)',
    '(C)',
    'OJSC',
    'PJSC',
    'S.p.A.',
    'Oyj',
    '(Unternehmen)',
    '(Company)',
]

blocklist_search = ['SDAX', 'MDAX', 'TecDAX', 'S&P 500', 'IBEX 35']

locations = {
    'OMX Helsinki 25': 'Finland',
    'SDAX': 'Germany',
    'MOEX': 'Russian Federation',
    'FTSE 100': 'United Kingdom',
    'S&P 100': 'United States',
    'IBEX 35': 'Finland',
    'DOW JONES': 'United States',
    'DAX': 'Germany',
    'CAC Mid 60': 'France',
    'TECDAX': 'Germany',
    'NASDAQ 100': 'United States',
    'Switzerland 20': 'Switzerland',
    'CAC 40': 'France',
    'S&P 500': 'United States',
    'OMX Stockholm 30': 'Sweden',
    'BEL 20': 'Belgium',
    'MDAX': 'Germany',
    'AEX': 'Netherlands',
}


def remove_most_common_endings(company):
    for end in most_common_endings:
        company = company.replace(f' {end}', '')
    return company.replace(',', '')


def long_to_short(name):
    result = name.replace('Aktiengesellschaft', 'AG')
    result = result.replace(', Inc.', ' Inc.')
    result = re.sub(r' \(.+\)$', '', result)
    if not result or not result.strip():
        raise RuntimeError(f'Error during name processing {name}')
    return result
